@@ -34,6 +34,9 @@ import atexit
 import gzip
 import shutil
 import tempfile
+# Import sleep directly into the module so we can monkey-patch it
+# during a test.
+from time import sleep
 from httplib2 import (
     FailedToDecompressContent, FileCache, Http, urlnorm)
 import simplejson
@@ -46,28 +49,6 @@ from lazr.uri import URI
 from errors import HTTPError
 from _json import DatetimeJSONEncoder
 
-# A drop-in replacement for httplib2's _decompressContent, which looks
-# in the Transfer-Encoding header instead of in Content-Encoding.
-def _decompressContent(response, new_content):
-    content = new_content
-    try:
-        encoding = response.get('transfer-encoding', None)
-        if encoding in ['gzip', 'deflate']:
-            if encoding == 'gzip':
-                content = gzip.GzipFile(
-                    fileobj=StringIO.StringIO(new_content)).read()
-            if encoding == 'deflate':
-                content = zlib.decompress(content)
-            response['content-length'] = str(len(content))
-            del response['transfer-encoding']
-    except IOError:
-        content = ""
-        raise FailedToDecompressContent(
-            ("Content purported to be compressed with %s but failed "
-             "to decompress." % response.get('transfer-encoding')),
-            response, content)
-    return content
-
 # A drop-in replacement for httplib2's safename.
 from httplib2 import _md5, re_url_scheme, re_slash
 def safename(filename):
@@ -96,15 +77,20 @@ def safename(filename):
     # filename is trimmed if it's longer than 200 characters, and then
     # a comma and a 32-character md5 sum are appended. This causes
     # problems on eCryptfs filesystems, where the maximum safe
-    # filename length is closer to 150 characters. So we take 117 as
-    # our limit (150-32-1) instead of 200.
+    # filename length is closer to 143 characters.
+    #
+    # We take a (user-hackable) maximum filename length from
+    # RestfulHttp and subtract 33 characters to make room for the comma
+    # and the md5 sum.
     #
     # See:
     #  http://code.google.com/p/httplib2/issues/detail?id=92
     #  https://bugs.launchpad.net/bugs/344878
-    #  https://bugs.launchpad.net/bugs/512832
-    if len(filename)>117:
-        filename=filename[:117]
+    #  https://bugs.launchpad.net/bugs/545197
+    maximum_filename_length = RestfulHttp.maximum_cache_filename_length
+    maximum_length_before_md5_sum = maximum_filename_length - 32 - 1
+    if len(filename) > maximum_length_before_md5_sum:
+        filename=filename[:maximum_length_before_md5_sum]
     return ",".join((filename, filemd5))
 
 
@@ -116,6 +102,8 @@ class RestfulHttp(Http):
     react when its cache is a MultipleRepresentationCache.
     """
 
+    maximum_cache_filename_length = 143
+
     def __init__(self, authorizer=None, cache=None, timeout=None,
                  proxy_info=None):
         super(RestfulHttp, self).__init__(cache, timeout, proxy_info)
@@ -125,26 +113,14 @@ class RestfulHttp(Http):
 
     def _request(self, conn, host, absolute_uri, request_uri, method, body,
                  headers, redirections, cachekey):
-        """Manipulate Transfer-Encoding header before sending the request.
-
-        httplib2 asks for compressed representations in
-        Accept-Encoding.  But a different content-encoding means a
-        different ETag, which can cause problems later when we make
-        a conditional request. We don't want to treat a
-        representation differently based on whether or not we asked
-        for a compressed version of it.
-
-        So we move the compression request from Accept-Encoding to
-        TE. Transfer-encoding compression can be handled
-        transparently, without affecting the ETag.
-        """
-        if 'accept-encoding' in headers:
-            headers['te'] = 'deflate, gzip'
-            del headers['accept-encoding']
+        """Use the authorizer to authorize an outgoing request."""
         if headers.has_key('authorization'):
             # There's an authorization header left over from a
-            # previous request that resulted in a redirect. Remove it
-            # and start again.
+            # previous request that resulted in a redirect. Resources
+            # protected by OAuth or HTTP Digest must send a distinct
+            # Authorization header with each request, to prevent
+            # playback attacks. Remove the Authorization header and
+            # start again.
             del headers['authorization']
         if self.authorizer is not None:
             self.authorizer.authorizeRequest(
@@ -153,15 +129,6 @@ class RestfulHttp(Http):
             conn, host, absolute_uri, request_uri, method, body, headers,
             redirections, cachekey)
 
-    def _conn_request(self, conn, request_uri, method, body, headers):
-        """Decompress content using our version of _decompressContent."""
-        response, content = super(RestfulHttp, self)._conn_request(
-            conn, request_uri, method, body, headers)
-        # Decompress the response, if it was compressed.
-        if method != "HEAD":
-            content = _decompressContent(response, content)
-        return (response, content)
-
     def _getCachedHeader(self, uri, header):
         """Retrieve a cached value for an HTTP header."""
         if isinstance(self.cache, MultipleRepresentationCache):
@@ -217,8 +184,11 @@ class MultipleRepresentationCache(FileCache):
 class Browser:
     """A class for making calls to lazr.restful web services."""
 
+    NOT_MODIFIED = object()
+    MAX_RETRIES = 6
+
     def __init__(self, service_root, credentials, cache=None, timeout=None,
-                 proxy_info=None):
+                 proxy_info=None, user_agent=None, max_retries=MAX_RETRIES):
         """Initialize, possibly creating a cache.
 
         If no cache is provided, a temporary directory will be used as
@@ -228,10 +198,27 @@ class Browser:
         if cache is None:
             cache = tempfile.mkdtemp()
             atexit.register(shutil.rmtree, cache)
-        if isinstance(cache, str):
+        if isinstance(cache, basestring):
             cache = MultipleRepresentationCache(cache)
         self._connection = service_root.httpFactory(
             credentials, cache, timeout, proxy_info)
+        self.user_agent = user_agent
+        self.max_retries = max_retries
+
+    def _request_and_retry(self, url, method, body, headers):
+        for retry_count in range(0, self.max_retries+1):
+            response, content = self._connection.request(
+                url, method=method, body=body, headers=headers)
+            if (response.status in [502, 503]
+                and retry_count < self.max_retries):
+                # The server returned a 502 or 503. Sleep for 0, 1, 2,
+                # 4, 8, 16, ... seconds and try again.
+                sleep_for = int(2**(retry_count-1))
+                sleep(sleep_for)
+            else:
+                break
+        # Either the request succeeded or we gave up.
+        return response, content
 
     def _request(self, url, data=None, method='GET',
                  media_type='application/json', extra_headers=None):
@@ -244,13 +231,47 @@ class Browser:
 
         # Add extra headers for the request.
         headers = {'Accept' : media_type}
+        if self.user_agent is not None:
+            headers['User-Agent'] = self.user_agent
         if isinstance(self._connection.cache, MultipleRepresentationCache):
             self._connection.cache.request_media_type = media_type
         if extra_headers is not None:
             headers.update(extra_headers)
         # Make the request.
-        response, content = self._connection.request(
+        response, content = self._request_and_retry(
             str(url), method=method, body=data, headers=headers)
+        if response.status == 304:
+            # The resource didn't change.
+            if content == '':
+                if ('If-None-Match' in headers
+                    or 'If-Modified-Since' in headers):
+                    # The caller made a conditional request, and the
+                    # condition failed. Rather than send an empty
+                    # representation, which might be misinterpreted,
+                    # send a special object that will let the calling code know
+                    # that the resource was not modified.
+                    return response, self.NOT_MODIFIED
+                else:
+                    # The caller didn't make a conditional request,
+                    # but the response code is 304 and there's no
+                    # content. The only way to handle this is to raise
+                    # an error.
+                    raise HTTPError(response, content)
+            else:
+                # XXX leonardr 2010/04/12 bug=httplib2#97
+                #
+                # Why is this check here? Why would there ever be any
+                # content when the response code is 304? It's because of
+                # an httplib2 bug that sometimes sets a 304 response
+                # code when caching retrieved documents. When the
+                # cached document is retrieved, we get a 304 response
+                # code and a full representation.
+                #
+                # Since the cache lookup succeeded, the 'real'
+                # response code is 200. This code undoes the bad
+                # behavior in httplib2.
+                response.status = 200
+            return response, content
         # Turn non-2xx responses into exceptions.
         if response.status // 100 != 2:
             raise HTTPError(response, content)
@@ -270,14 +291,8 @@ class Browser:
 
     def get_wadl_application(self, url):
         """GET a WADL representation of the resource at the requested url."""
-        # We're probably talking to an old version of lazr.restful
-        # that misspells the WADL media type. Accept either the correctly
-        # spelled media type or the misspelling.
         wadl_type = 'application/vnd.sun.wadl+xml'
-        misspelled_wadl_type = 'application/vd.sun.wadl+xml'
-        accept = "%s, %s" % (wadl_type, misspelled_wadl_type)
-        response, content = self._request(
-            url, media_type=wadl_type, extra_headers={'Accept': accept})
+        response, content = self._request(url, media_type=wadl_type)
         return Application(str(url), content)
 
     def post(self, url, method_name, **kws):
@@ -297,6 +312,7 @@ class Browser:
     def delete(self, url):
         """DELETE the resource at the given URL."""
         self._request(url, method='DELETE')
+        return None
 
     def patch(self, url, representation, headers=None):
         """PATCH the object at url with the updated representation."""
