@@ -30,10 +30,11 @@ __all__ = [
 
 
 import cgi
+from email.message import Message
 import simplejson
 from StringIO import StringIO
 import urllib
-from urlparse import urlparse
+from urlparse import urlparse, urljoin
 
 from lazr.uri import URI
 from wadllib.application import Resource as WadlResource
@@ -41,6 +42,8 @@ from _browser import Browser, RestfulHttp
 from _json import DatetimeJSONEncoder
 from errors import HTTPError
 
+from lazr.restfulclient import __version__
+
 missing = object()
 
 class HeaderDictionary:
@@ -282,15 +285,11 @@ class Resource(RestfulBase):
         headers = {}
         if etag is not None:
             headers['If-None-Match'] = etag
-        try:
-            representation = self._root._browser.get(
-                self._wadl_resource, headers=headers)
-        except HTTPError, e:
-            if e.response['status'] == '304':
-                # The entry wasn't modified. No need to do anything.
-                return
-            else:
-                raise e
+        representation = self._root._browser.get(
+            self._wadl_resource, headers=headers)
+        if representation == self._root._browser.NOT_MODIFIED:
+            # The entry wasn't modified. No need to do anything.
+            return
         # __setattr__ assumes we're setting an attribute of the resource,
         # so we manipulate __dict__ directly.
         self.__dict__['_wadl_resource'] = self._wadl_resource.bind(
@@ -299,11 +298,11 @@ class Resource(RestfulBase):
     def __getattr__(self, attr):
         """Try to retrive a named operation or parameter of the given name."""
         try:
-            return self.lp_get_parameter(attr)
+            return self.lp_get_named_operation(attr)
         except KeyError:
             pass
         try:
-            return self.lp_get_named_operation(attr)
+            return self.lp_get_parameter(attr)
         except KeyError:
             raise AttributeError("'%s' object has no attribute '%s'"
                                  % (self.__class__.__name__, attr))
@@ -330,8 +329,19 @@ class Resource(RestfulBase):
         if self._wadl_resource.representation is None:
             # Get a representation of the linked resource.
             representation = self._root._browser.get(self._wadl_resource)
+            representation = simplejson.loads(representation)
+            type_link = representation['resource_type_link']
+            if (type_link is not None
+                and type_link != self._wadl_resource.type_url):
+                # In rare cases, the resource type served by the
+                # server conflicts with the type the client thought
+                # this resource had. When this happens, the server
+                # value takes precedence.
+                resource_type = self._root._wadl.get_resource_type(type_link)
+                self._wadl_resource.tag = resource_type.tag
             self.__dict__['_wadl_resource'] = self._wadl_resource.bind(
-                representation, self.JSON_MEDIA_TYPE)
+                representation, self.JSON_MEDIA_TYPE,
+                representation_needs_processing=False)
 
     def __ne__(self, other):
         """Inequality operator."""
@@ -380,7 +390,8 @@ class ServiceRoot(Resource):
     RESOURCE_TYPE_CLASSES = {'HostedFile': HostedFile}
 
     def __init__(self, authorizer, service_root, cache=None,
-                 timeout=None, proxy_info=None, version=None):
+                 timeout=None, proxy_info=None, version=None,
+                 base_client_name='', max_retries=Browser.MAX_RETRIES):
         """Root access to a lazr.restful API.
 
         :param credentials: The credentials used to access the service.
@@ -394,9 +405,15 @@ class ServiceRoot(Resource):
             if service_root[-1] != '/':
                 service_root += '/'
         self._root_uri = URI(service_root)
+
+        # Set up data necessary to calculate the User-Agent header.
+        self._base_client_name = base_client_name
+
         # Get the WADL definition.
+        self.credentials = authorizer
         self._browser = Browser(
-            self, authorizer, cache, timeout, proxy_info)
+            self, authorizer, cache, timeout, proxy_info, self._user_agent,
+            max_retries)
         self._wadl = self._browser.get_wadl_application(self._root_uri)
 
         # Get the root resource.
@@ -404,13 +421,42 @@ class ServiceRoot(Resource):
         bound_root = root_resource.bind(
             self._browser.get(root_resource), 'application/json')
         super(ServiceRoot, self).__init__(None, bound_root)
-        self.credentials = authorizer
+
+    @property
+    def _user_agent(self):
+        """The value for the User-Agent header.
+
+        This will be something like:
+        launchpadlib 1.6.1, lazr.restfulclient 1.0.0; oauth_consumer=apport
+
+        That is, a string describing lazr.restfulclient and an
+        optional custom client built on top, and parameters
+        containing any authorization-specific information that
+        identifies the user agent (such as the OAuth consumer key).
+        """
+        base_portion = "lazr.restfulclient %s" % __version__
+        if self._base_client_name != '':
+            base_portion = self._base_client_name + ' (' + base_portion + ')'
+
+        message = Message()
+        message['User-Agent'] = base_portion
+        if self.credentials is not None:
+            for key, value in self.credentials.user_agent_params.items():
+                message.set_param(key, value, 'User-Agent')
+        return message['User-Agent']
 
     def httpFactory(self, authorizer, cache, timeout, proxy_info):
         return RestfulHttp(authorizer, cache, timeout, proxy_info)
 
     def load(self, url):
         """Load a resource given its URL."""
+        parsed = urlparse(url)
+        if parsed.scheme == '':
+            # This is a relative URL. Make it absolute by joining
+            # it with the service root resource.
+            if url[:1] == '/':
+                url = url[1:]
+            url = self._root_uri.append(url)
         document = self._browser.get(url)
         try:
             representation = simplejson.loads(unicode(document))
@@ -581,6 +627,10 @@ class Entry(Resource):
         return '<%s at %s>' % (
             URI(self.resource_type_link).fragment, self.self_link)
 
+    def lp_delete(self):
+        """Delete the resource."""
+        return self._root._browser.delete(URI(self.self_link))
+
     def __str__(self):
         """Return the URL to the resource."""
         return self.self_link
@@ -845,28 +895,43 @@ class CollectionWithKeyBasedLookup(Collection):
         if url is None:
             raise KeyError(key)
 
-        # We don't know what kind of resource this is. Even the
-        # subclass doesn't necessarily know, because some resources
-        # (the person list) are gateways to more than one kind of
-        # resource (people, and teams). The only way to know for sure
-        # is to retrieve a representation of the resource and see how
-        # the resource describes itself.
-        try:
-            representation = simplejson.loads(
-                unicode(self._root._browser.get(url)))
-        except HTTPError, error:
-            # There's no resource corresponding to the given ID.
-            if error.response.status == 404:
-                raise KeyError(key)
-            raise
-        # We know that every lazr.restful resource has a
-        # 'resource_type_link' in its representation.
-        resource_type_link = representation['resource_type_link']
+        if self.collection_of is not None:
+            # We know what kind of resource is at the other end of the
+            # URL. There's no need to actually fetch that URL until
+            # the user demands it. If the user is invoking a named
+            # operation on this object rather than fetching its data,
+            # this will save us one round trip.
+            representation = None
+            resource_type_link = urljoin(
+                self._root._wadl.markup_url, '#' + self.collection_of)
+        else:
+            # We don't know what kind of resource this is. Either the
+            # subclass wasn't programmed with this knowledge, or
+            # there's simply no way to tell without going to the
+            # server, because the collection contains more than one
+            # kind of resource. The only way to know for sure is to
+            # retrieve a representation of the resource and see how
+            # the resource describes itself.
+            try:
+                representation = simplejson.loads(
+                    unicode(self._root._browser.get(url)))
+            except HTTPError, error:
+                # There's no resource corresponding to the given ID.
+                if error.response.status == 404:
+                    raise KeyError(key)
+                raise
+            # We know that every lazr.restful resource has a
+            # 'resource_type_link' in its representation.
+            resource_type_link = representation['resource_type_link']
+
         resource = WadlResource(self._root._wadl, url, resource_type_link)
         return self._create_bound_resource(
             self._root, resource, representation=representation,
             representation_needs_processing=False)
 
+    # If provided, this should be a string designating the ID of a
+    # resource_type from a specific service's WADL file.
+    collection_of = None
 
     def _get_url_from_id(self, key):
         """Transform the unique ID of an object into its URL."""
